"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import asyncio
import inspect
import logging
import random
from collections import defaultdict
from inspect import getmembers, ismethod
from typing import (Any, Callable, Coroutine, Dict, Generic, List, Optional, Sequence, Set, Tuple,
                    Type, TypeVar, Union)

import aiohttp

from .abc import BasePlayer, Source
from .events import Event
from .node import Node
from .nodemanager import NodeManager
from .player import DefaultPlayer
from .playermanager import PlayerManager
from .server import AudioTrack, LoadResult

_log = logging.getLogger(__name__)

PlayerT = TypeVar('PlayerT', bound=BasePlayer)
EventT = TypeVar('EventT', bound=Event)


class Client(Generic[PlayerT]):
    """
    Represents a Lavalink client used to manage nodes and connections.

    Parameters
    ----------
    user_id: Union[:class:`int`, :class:`str`]
        The user id of the bot.
    player: Type[:class:`BasePlayer`]
        The class that should be used for the player. Defaults to :class:`DefaultPlayer`.
        Do not change this unless you know what you are doing!
    regions: Optional[Dict[str, Tuple[str]]]
        A mapping of continent -> Discord RTC regions.
        The key should be an identifier used when instantiating an node.
        The values should be a list of RTC regions that will be handled by the associated identifying key.

        Example: ``{"us": ("us-central", "us-east", "us-south", "us-west", "brazil")}``

        You should only change this if you know what you're doing and want more control over
        region groups. Defaults to ``None``.
    connect_back: Optional[:class:`bool`]
        A boolean that determines if a player will connect back to the
        node it was originally connected to. This is not recommended to do since
        the player will most likely be performing better in the new node. Defaults to ``False``.

        Warning
        -------
        If this option is enabled and the player's node is changed through `Player.change_node` after
        the player was moved via the failover mechanism, the player will still move back to the original
        node when it becomes available. This behaviour can be avoided in custom player implementations by
        setting ``self._original_node`` to ``None`` in the :func:`BasePlayer.change_node` function.
    request_timeout: Optional[:class:`aiohttp.ClientTimeout`]
        A ``ClientTimeout`` object to be used for requests. If unspecified, defaults will be used.

    Attributes
    ----------
    node_manager: :class:`NodeManager`
        The node manager, used for storing and managing all registered Lavalink nodes.
    player_manager: :class:`PlayerManager`
        The player manager, used for storing and managing all players.
    sources: Set[:class:`Source`]
        The custom sources registered to this client.
    """
    __slots__ = ('_session', '_user_id', '_event_hooks', 'node_manager', 'player_manager', 'sources')

    def __init__(self, user_id: Union[int, str], player: Type[PlayerT] = DefaultPlayer,
                 regions: Optional[Dict[str, Tuple[str]]] = None, connect_back: bool = False,
                 request_timeout: Optional[aiohttp.ClientTimeout] = None):
        if not isinstance(user_id, (str, int)) or isinstance(user_id, bool):
            # bool has special handling because it subclasses `int`, so will return True for the first isinstance check.
            raise TypeError(f'user_id must be either an int or str (not {type(user_id).__name__}). If the type is None, '
                            'ensure your bot has fired "on_ready" before instantiating '
                            'the Lavalink client. Alternatively, you can hardcode your user ID.')

        self._session = aiohttp.ClientSession(timeout=request_timeout)
        self._user_id = int(user_id)
        self._event_hooks = defaultdict(list)
        self.node_manager: NodeManager = NodeManager(self, regions, connect_back)
        self.player_manager: PlayerManager[PlayerT] = PlayerManager(self, player)
        self.sources: Set[Source] = set()

    @property
    def nodes(self) -> List[Node]:
        """
        Convenience shortcut for :attr:`NodeManager.nodes`.
        """
        return self.node_manager.nodes

    @property
    def players(self) -> Dict[int, PlayerT]:
        """
        Convenience shortcut for :attr:`PlayerManager.players`.
        """
        return self.player_manager.players

    async def close(self):
        """|coro|

        Closes all active connections and frees any resources in use.
        """
        for node in self.node_manager:
            await node.destroy()

        await self._session.close()

    def add_event_hook(self, *hooks, event: Optional[Type[EventT]] = None):
        """
        Adds one or more event hooks to be dispatched on an event.

        Note
        ----
        Track event dispatch order is not guaranteed!
        For example, this means you could receive a :class:`TrackStartEvent` before you receive a
        :class:`TrackEndEvent` when executing operations such as ``skip()``.

        Parameters
        ----------
        hooks: :class:`function`
            The hooks to register for the given event type.
            If ``event`` parameter is left empty, then it will run when any event is dispatched.
        event: Optional[Type[:class:`Event`]]
            The event the hooks belong to. They will be called when that specific event type is
            dispatched. Defaults to ``None`` which means the hook is dispatched on all events.
        """
        if event is not None and Event not in event.__bases__:
            raise TypeError('Event parameter is not of type Event or None')

        event_name = event.__name__ if event is not None else 'Generic'
        event_hooks = self._event_hooks[event_name]

        for hook in hooks:
            if not callable(hook) or not inspect.iscoroutinefunction(hook):
                raise TypeError('Hook is not callable or a coroutine')

            if hook not in event_hooks:
                event_hooks.append(hook)

    def add_event_hooks(self, cls: object):
        """
        Scans the provided class ``cls`` for functions decorated with :func:`listener`,
        and sets them up to process Lavalink events.

        Example:

            .. code:: python

                # Inside a class __init__ method
                self.client = lavalink.Client(...)
                self.client.add_event_hooks(self)

        Note
        ----
        Track event dispatch order is not guaranteed!
        For example, this means you could receive a :class:`TrackStartEvent` before you receive a
        :class:`TrackEndEvent` when executing operations such as ``skip()``.

        Parameters
        ----------
        cls: object
            An instance of a class containing event hook methods.
        """
        methods = getmembers(cls, predicate=lambda meth: hasattr(meth, '__name__')
                             and not meth.__name__.startswith('_') and ismethod(meth)
                             and hasattr(meth, '_lavalink_events'))

        for _, listener in methods:  # _ = meth_name
            # wrapped = partial(listener, cls)
            events = listener._lavalink_events

            if events:
                for event in events:
                    self._event_hooks[event.__name__].append(listener)
            else:
                self._event_hooks['Generic'].append(listener)

    def remove_event_hooks(self, *, events: Optional[Sequence[Type[EventT]]] = None, hooks: Sequence[Callable]):
        """
        Removes the given hooks from the event hook registry.

        Parameters
        ----------
        events: Sequence[:class:`Event`]
            The events to remove the hooks from. This parameter can be omitted,
            and the events registered on the function via :func:`listener` will be used instead, if applicable.
            Otherwise, a default value of ``Generic`` is used instead.
        hooks: Sequence[Callable]
            A list of hook methods to remove.
        """
        if events is not None:
            for event in events:
                if Event not in event.__bases__:
                    raise TypeError(f'{event.__name__} is not of type Event')

        for hook in hooks:
            if not callable(hook):
                raise ValueError(f'Provided hook {hook} is not a callable')

        for hook in hooks:
            unregister_events = events or getattr(hook, '_lavalink_events', None)

            try:
                if not unregister_events:
                    self._event_hooks['Generic'].remove(hook)
                else:
                    for event in unregister_events:
                        self._event_hooks[event.__name__].remove(hook)
            except ValueError:
                pass

    def register_source(self, source: Source):
        """
        Registers a :class:`Source` that Lavalink.py will use for looking up tracks.

        Parameters
        ----------
        source: :class:`Source`
            The source to register.
        """
        if not isinstance(source, Source):
            raise TypeError(f'Class \'{type(source).__name__}\' must inherit Source!')

        self.sources.add(source)

    def get_source(self, source_name: str) -> Optional[Source]:
        """
        Gets a registered source by the given name.

        Parameters
        ----------
        source_name: :class:`str`
            The name of the source to get.

        Returns
        -------
        Optional[:class:`Source`]
            The source with the matching name. May be ``None`` if
            the name didn't match any of those in the registered sources.
        """
        return next((source for source in self.sources if source.name == source_name), None)

    def add_node(self, host: str, port: int, password: str, region: str, name: Optional[str] = None,
                 ssl: bool = False, session_id: Optional[str] = None, connect: bool = True, tags: Optional[Dict[str, Any]] = None) -> Node:
        """
        Shortcut for :func:`NodeManager.add_node`.

        Adds a node to Lavalink's node manager.

        Parameters
        ----------
        host: :class:`str`
            The address of the Lavalink node.
        port: :class:`int`
            The port to use for websocket and REST connections.
        password: :class:`str`
            The password used for authentication.
        region: :class:`str`
            The region to assign this node to.
        name: Optional[:class:`str`]
            An identifier for the node that will show in logs. Defaults to ``None``.
        ssl: :class:`bool`
            Whether to use SSL for the node. SSL will use ``wss`` and ``https``, instead of ``ws`` and ``http``,
            respectively. Your node should support SSL if you intend to enable this, either via reverse proxy or
            other methods. Only enable this if you know what you're doing.
        session_id: Optional[:class:`str`]
            The ID of the session to resume. Defaults to ``None``.
            Only specify this if you have the ID of the session you want to resume.
        connect: :class:`bool`
            Whether to immediately connect to the node after creating it.
            If ``False``, you must call :func:`Node.connect` if you require WebSocket functionality.
        tags: Optional[Dict[:class:`str`, Any]]
            Additional tags to attach to this node. You can use this to store additional metadata
            that you may need to access later.

        Returns
        -------
        :class:`Node`
            The created Node instance.
        """
        return self.node_manager.add_node(host, port, password, region, name, ssl,
                                          session_id, connect, tags)

    async def get_local_tracks(self, query: str) -> LoadResult:
        """|coro|

        Searches :attr:`sources` registered to this client for the given query.

        Parameters
        ----------
        query: :class:`str`
            The query to perform a search for.

        Returns
        -------
        :class:`LoadResult`
        """
        for source in self.sources:
            load_result = await source.load_item(self, query)

            if load_result:
                return load_result

        return LoadResult.empty()

    async def get_tracks(self, query: str, node: Optional[Node] = None,
                         check_local: bool = False) -> LoadResult:
        """|coro|

        Retrieves a list of results pertaining to the provided query.

        If ``check_local`` is set to ``True`` and any of the sources return a :class:`LoadResult`
        then that result will be returned, and Lavalink will not be queried.

        Warning
        -------
        Avoid setting ``check_local`` to ``True`` if you call this method from a custom :class:`Source` to avoid
        recursion issues!

        Parameters
        ----------
        query: :class:`str`
            The query to perform a search for.
        node: Optional[:class:`Node`]
            The node to use for track lookup. Leave this blank to use a random node.
            Defaults to ``None`` which is a random node.
        check_local: :class:`bool`
            Whether to also search the query on sources registered with this Lavalink client.

        Returns
        -------
        :class:`LoadResult`
        """
        if check_local:
            for source in self.sources:
                load_result = await source.load_item(self, query)

                if load_result:
                    return load_result

        node = node or random.choice(self.node_manager.available_nodes) or random.choice(self.node_manager.nodes)
        return await node.get_tracks(query)

    async def decode_track(self, track: str, node: Optional[Node] = None) -> AudioTrack:
        """|coro|

        Decodes a base64-encoded track string into a dict.

        Parameters
        ----------
        track: :class:`str`
            The base64-encoded ``track`` string.
        node: Optional[:class:`Node`]
            The node to use for the query. Defaults to ``None`` which is a random node.

        Returns
        -------
        :class:`AudioTrack`
        """
        node = node or random.choice(self.node_manager.available_nodes) or random.choice(self.node_manager.nodes)
        return await node.decode_track(track)

    async def decode_tracks(self, tracks: List[str], node: Optional[Node] = None) -> List[AudioTrack]:
        """|coro|

        Decodes a list of base64-encoded track strings into a list of :class:`AudioTrack`.

        Parameters
        ----------
        tracks: List[:class:`str`]
            A list of base64-encoded ``track`` strings.
        node: Optional[:class:`Node`]
            The node to use for the query. Defaults to ``None`` which is a random node.

        Returns
        -------
        List[:class:`AudioTrack`]
            A list of decoded :class:`AudioTrack`.
        """
        node = node or random.choice(self.node_manager.available_nodes) or random.choice(self.node_manager.nodes)
        return await node.decode_tracks(tracks)

    async def voice_update_handler(self, data: Dict[str, Any]):
        """|coro|

        This function intercepts websocket data from your Discord library and
        forwards the relevant information on to Lavalink, which is used to
        establish a websocket connection and send audio packets to Discord.

        Example
        -------
        .. code:: python

            bot.add_listener(lavalink_client.voice_update_handler, 'on_socket_response')

        Parameters
        ----------
        data: Dict[str, Any]
            The payload received from Discord.
        """
        if not data or 't' not in data:
            return

        if data['t'] == 'VOICE_SERVER_UPDATE':
            guild_id = int(data['d']['guild_id'])
            player = self.player_manager.get(guild_id)

            if player:
                await player._voice_server_update(data['d'])
        elif data['t'] == 'VOICE_STATE_UPDATE':
            if int(data['d']['user_id']) != self._user_id:
                return

            guild_id = int(data['d']['guild_id'])
            player = self.player_manager.get(guild_id)

            if player:
                await player._voice_state_update(data['d'])

    def has_listeners(self, event: Type[Event]) -> bool:
        """
        Check whether the client has any listeners for a specific event type.
        """
        return len(self._event_hooks['Generic']) > 0 or len(self._event_hooks[event.__name__]) > 0

    def _dispatch_event(self, event: Event):
        """
        Dispatches the given event to all registered hooks.

        Parameters
        ----------
        event: :class:`Event`
            The event to dispatch to the hooks.
        """
        generic_hooks = self._event_hooks['Generic']
        targeted_hooks = self._event_hooks[type(event).__name__]

        if not generic_hooks and not targeted_hooks:
            return

        hooks = generic_hooks + targeted_hooks
        loop = asyncio.get_event_loop()
        loop.create_task(self.__real_dispatch(event, hooks))

    async def __real_dispatch(self, event: Event, hooks: List[Callable[[Event], Coroutine[Any, Any, Any]]]):
        async def _hook_wrapper(hook, event):
            try:
                await hook(event)
            except asyncio.CancelledError:
                _log.debug('Event hook \'%s\' dispatch was cancelled.', hook.__name__)
            except:  # noqa: E722 pylint: disable=bare-except
                _log.exception('Event hook \'%s\' encountered an exception!', hook.__name__)

        tasks = [_hook_wrapper(hook, event) for hook in hooks]
        await asyncio.gather(*tasks)
        _log.debug('Dispatched \'%s\' to all registered hooks', type(event).__name__)

    def __repr__(self):
        return f'<Client user_id={self._user_id} nodes={len(self.node_manager)} players={len(self.player_manager)}>'



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from asyncio import Task
from collections import defaultdict
from time import time
from typing import (TYPE_CHECKING, Any, Dict, List, Optional, Type, TypeVar,
                    Union, cast, overload)

from .abc import BasePlayer, Filter
from .common import MISSING
from .errors import AuthenticationError, ClientError, RequestError
from .server import (AudioTrack, LoadResult, RawInfo, RawPlayer,
                     RawPlayerVoiceState, RawRouteplanner, RawSession,
                     RawStats)
from .stats import Stats
from .transport import Transport

if TYPE_CHECKING:
    from .client import Client
    from .nodemanager import NodeManager

T = TypeVar('T')


class Node:
    """
    Represents a Node connection with Lavalink.

    Note
    ----
    To construct a node, you should use :func:`Client.add_node` instead.

    Attributes
    ----------
    client: :class:`Client`
        The Lavalink client.
    region: :class:`str`
        The region to assign this node to.
    name: :class:`str`
        The name the :class:`Node` is identified by.
    stats: :class:`Stats`
        The statistics of how the :class:`Node` is performing.
    tags: Dict[:class:`str`, Any]
        Additional tags to attach to this node.
    """
    __slots__ = ('client', 'manager', '_transport', 'region', 'name', 'stats', 'tags')

    def __init__(self, manager, host: str, port: int, password: str, region: str, name: Optional[str] = None,
                 ssl: bool = False, session_id: Optional[str] = None, connect: bool = True, tags: Optional[Dict[str, Any]] = None):
        self.client: 'Client' = manager.client
        self.manager: 'NodeManager' = manager
        self._transport = Transport(self, host, port, password, ssl, session_id, connect)

        self.region: str = region
        self.name: str = name or f'{region}-{host}:{port}'
        self.stats: Stats = Stats.empty(self)
        self.tags: Dict[str, Any] = tags or {}

    @property
    def session_id(self) -> Optional[str]:
        """
        The session ID for this node.
        Could be ``None`` if a ready event has not yet been received from the server.
        """
        return self._transport.session_id

    @property
    def available(self) -> bool:
        """
        Returns whether the node has a websocket connection.
        The node could *probably* still be used for HTTP requests even without a WS connection.
        """
        return self._transport.ws_connected

    @property
    def _original_players(self) -> List[BasePlayer]:
        """
        Returns a list of players that were assigned to this node, but were moved due to failover etc.

        Returns
        -------
        List[:class:`BasePlayer`]
        """
        return [p for p in self.client.player_manager.values() if p._original_node == self]

    @property
    def players(self) -> List[BasePlayer]:
        """
        Returns a list of all players on this node.

        Returns
        -------
        List[:class:`BasePlayer`]
        """
        return [p for p in self.client.player_manager.values() if p.node == self]

    @property
    def penalty(self) -> float:
        """ Returns the load-balancing penalty for this node. """
        if not self.available or not self.stats:
            return 9e30

        return self.stats.penalty.total

    async def get_rest_latency(self) -> float:
        """|coro|

        Measures the REST latency for this node.
        This simply calls :func:`get_version` but measures the time between when the request was made,
        and when a response was received.

        Returns
        -------
        float
            The latency, in milliseconds. ``-1`` if an error occurred during the request (e.g. node is unreachable),
            otherwise, a positive number.
        """
        start = time()

        try:
            await self.get_version()
        except (AuthenticationError, ClientError, RequestError):
            return -1

        return (time() - start) * 1000

    async def connect(self, force: bool = False) -> Optional[Task]:
        """|coro|

        Initiates a WebSocket connection to this node.
        If a connection already exists, and ``force`` is ``False``, this will not do anything.

        Parameters
        ----------
        force: :class:`bool`
            Whether to close any existing WebSocket connections and re-establish a connection to
            the node.

        Returns
        -------
        Optional[:class:`asyncio.Task`]
            The WebSocket connection task, or ``None`` if a WebSocket connection already exists and force
            is ``False``.
        """
        if self._transport.ws_connected:
            if not force:
                return None

            await self._transport.close()

        return self._transport.connect()

    async def destroy(self):
        """|coro|

        Destroys the transport and any underlying connections for this node.
        This will also cleanly close the websocket.
        """
        await self._transport.destroy()

    async def get_tracks(self, query: str) -> LoadResult:
        """|coro|

        Retrieves a list of results pertaining to the provided query.

        Parameters
        ----------
        query: :class:`str`
            The query to perform a search for.

        Returns
        -------
        :class:`LoadResult`
        """
        return await self.request('GET', 'loadtracks', params={'identifier': query}, to=LoadResult)

    async def decode_track(self, track: str) -> AudioTrack:
        """|coro|

        Decodes a base64-encoded track string into an :class:`AudioTrack` object.

        Parameters
        ----------
        track: :class:`str`
            The base64-encoded track string to decode.

        Returns
        -------
        :class:`AudioTrack`
        """
        return await self.request('GET', 'decodetrack', params={'track': track}, to=AudioTrack)

    async def decode_tracks(self, tracks: List[str]) -> List[AudioTrack]:
        """|coro|

        Decodes a list of base64-encoded track strings into a list of :class:`AudioTrack`.

        Parameters
        ----------
        tracks: List[:class:`str`]
            A list of base64-encoded ``track`` strings.

        Returns
        -------
        List[:class:`AudioTrack`]
            A list of decoded AudioTracks.
        """
        response = await self.request('POST', 'decodetracks', json=tracks)
        return list(map(AudioTrack, response))  # type: ignore

    async def get_routeplanner_status(self) -> Optional[RawRouteplanner]:
        """|coro|

        Retrieves the status of the target node's routeplanner.

        Returns
        -------
        Optional[:class:`RawRouteplanner`]
            A dict representing the routeplanner information, or ``None`` if
            the routeplanner is not enabled on the server.
        """
        response = await self.request('GET', 'routeplanner/status')

        if response is True:  # HTTP status 204
            return None

        return cast(RawRouteplanner, response)

    async def routeplanner_free_address(self, address: str) -> bool:
        """|coro|

        Frees up the provided IP address in the target node's routeplanner.

        Parameters
        ----------
        address: :class:`str`
            The address to free.

        Returns
        -------
        :class:`bool`
            True if the address was freed, False otherwise.
        """
        try:
            return await self.request('POST', 'routeplanner/free/address', json={'address': address})  # type: ignore
        except RequestError:
            return False

    async def routeplanner_free_all_failing(self) -> bool:
        """|coro|

        Frees up all IP addresses in the target node that have been marked as failing.

        Returns
        -------
        :class:`bool`
            True if all failing addresses were freed, False otherwise.
        """
        try:
            return await self.request('POST', 'routeplanner/free/all')  # type: ignore
        except RequestError:
            return False

    async def get_info(self) -> RawInfo:
        """|coro|

        Retrieves information about this node.

        Returns
        -------
        :class:`RawInfo`
            A raw response containing information about the node.
        """
        return cast(RawInfo, await self.request('GET', 'info'))

    async def get_stats(self) -> RawStats:
        """|coro|

        Retrieves statistics about this node.

        Returns
        -------
        :class:`RawStats`
            A raw response containing information about the node.
        """
        return cast(RawStats, await self.request('GET', 'stats'))

    async def get_version(self) -> str:
        """|coro|

        Retrieves the version of this node.

        Returns
        -------
        str
            The version of this Lavalink server.
        """
        return await self.request('GET', 'version', to=str, versioned=False)

    async def get_player(self, guild_id: Union[str, int]) -> RawPlayer:
        """|coro|

        Retrieves a player from the node.
        This returns raw data, to retrieve a player you can interact with, use :func:`PlayerManager.get`.

        Returns
        -------
        :class:`RawPlayer`
        """
        session_id = self.session_id

        if not session_id:
            raise ClientError('Cannot retrieve a player without a valid session ID!')

        return await self.request('GET', f'sessions/{session_id}/players/{guild_id}')  # type: ignore

    async def get_players(self) -> List[RawPlayer]:
        """|coro|

        Retrieves a list of players from the node.
        This returns raw data, to retrieve players you can interact with, use :attr:`players`.

        Returns
        -------
        List[:class:`RawPlayer`]
        """
        session_id = self.session_id

        if not session_id:
            raise ClientError('Cannot retrieve a list of players without a valid session ID!')

        return await self.request('GET', f'sessions/{session_id}/players')  # type: ignore

    @overload
    async def update_player(self,
                            *,
                            guild_id: Union[str, int],
                            encoded_track: Optional[str] = ...,
                            no_replace: bool = ...,
                            position: int = ...,
                            end_time: int = ...,
                            volume: int = ...,
                            paused: bool = ...,
                            filters: Optional[List[Filter]] = ...,
                            voice_state: RawPlayerVoiceState = ...,
                            user_data: Dict[str, Any] = ...,
                            **kwargs) -> Optional[RawPlayer]:
        ...

    @overload
    async def update_player(self,
                            *,
                            guild_id: Union[str, int],
                            identifier: str = ...,
                            no_replace: bool = ...,
                            position: int = ...,
                            end_time: int = ...,
                            volume: int = ...,
                            paused: bool = ...,
                            filters: Optional[List[Filter]] = ...,
                            voice_state: RawPlayerVoiceState = ...,
                            user_data: Dict[str, Any] = ...,
                            **kwargs) -> Optional[RawPlayer]:
        ...

    @overload
    async def update_player(self,
                            *,
                            guild_id: Union[str, int],
                            no_replace: bool = ...,
                            position: int = ...,
                            end_time: int = ...,
                            volume: int = ...,
                            paused: bool = ...,
                            filters: Optional[List[Filter]] = ...,
                            voice_state: RawPlayerVoiceState = ...,
                            user_data: Dict[str, Any] = ...,
                            **kwargs) -> Optional[RawPlayer]:
        ...

    async def update_player(self,  # pylint: disable=too-many-locals
                            guild_id: Union[str, int],
                            encoded_track: Optional[str] = MISSING,
                            identifier: str = MISSING,
                            no_replace: bool = MISSING,
                            position: int = MISSING,
                            end_time: int = MISSING,
                            volume: int = MISSING,
                            paused: bool = MISSING,
                            filters: Optional[List[Filter]] = MISSING,
                            voice_state: RawPlayerVoiceState = MISSING,
                            user_data: Dict[str, Any] = MISSING,
                            **kwargs) -> Optional[RawPlayer]:
        """|coro|

        .. _response object: https://lavalink.dev/api/rest#Player

        Update the state of a player.

        Warning
        -------
        If this function is called directly, rather than through, e.g. a player,
        the internal state is not guaranteed! This means that any attributes accessible through other classes
        may not correspond with those stored in, or provided by the server. Use with caution!

        Parameters
        ----------
        guild_id: Union[str, int]
            The guild ID of the player to update.
        encoded_track: Optional[str]
            The base64-encoded track string to play.
            You may provide ``None`` to stop the player.

            Warning
            -------
            This option is mutually exclusive with ``identifier``. You cannot provide both options.
        identifier: str
            The identifier of the track to play. This can be a track ID or URL. It may not be a
            search query or playlist link. If it yields a search, playlist, or no track, a :class:`RequestError`
            will be raised.

            Warning
            -------
            This option is mutually exclusive with ``encoded_track``. You cannot provide both options.
        no_replace: bool
            Whether to replace the currently playing track (if one exists) with the new track.
            Only takes effect if ``identifier`` or ``encoded_track`` is provided.
            This parameter will only take effect when a track is provided.
        position: int
            The track position in milliseconds. This can be used to seek.
        end_time: int
            The position, in milliseconds, to end the track at.
        volume: int
            The new volume of the player. This must be within the range of 0 to 1000.
        paused: bool
            Whether to pause the player.
        filters: Optional[List[:class:`Filter`]]
            The filters to apply to the player.
            Specify ``None`` or ``[]`` to clear.
        voice_state: :class:`RawPlayerVoiceState`
            The new voice state of the player.
        user_data: Dict[str, Any]
            The user data to attach to the track, if one is provided.
            This parameter will only take effect when a track is provided.
        **kwargs: Any
            The kwargs to use when updating the player. You can specify any extra parameters that may be
            used by plugins, which offer extra features not supported out-of-the-box by Lavalink.py.

        Returns
        -------
        Optional[:class:`RawPlayer`]
            The raw player update `response object`_, or ``None`` , if a request wasn't made due to an
            empty payload.
        """
        session_id = self.session_id

        if not session_id:
            raise ClientError('Cannot update the state of a player without a valid session ID!')

        if encoded_track is not MISSING and identifier is not MISSING:
            raise ValueError('encoded_track and identifier are mutually exclusive options, you may not specify both together.')

        params = {}
        json = kwargs

        if identifier is not MISSING or encoded_track is not MISSING:
            track = {}

            if identifier is not MISSING:
                track['identifier'] = identifier
            elif encoded_track is not MISSING:
                track['encoded'] = encoded_track

            if user_data is not MISSING:
                track['userData'] = user_data

            if no_replace is not MISSING:
                params['noReplace'] = str(no_replace).lower()

            json['track'] = track

        if position is not MISSING:
            if not isinstance(position, (int, float)):
                raise ValueError('position must be an int!')

            json['position'] = position

        if end_time is not MISSING:
            if not isinstance(end_time, int) or end_time <= 0:
                raise ValueError('end_time must be an int, and greater than 0!')

            json['endTime'] = end_time

        if volume is not MISSING:
            if not isinstance(volume, int) or not 0 <= volume <= 1000:
                raise ValueError('volume must be an int, and within the range of 0 to 1000!')

            json['volume'] = volume

        if paused is not MISSING:
            if not isinstance(paused, bool):
                raise ValueError('paused must be a bool!')

            json['paused'] = paused

        if filters is not MISSING:
            if filters is not None:
                if not isinstance(filters, list) or not all(isinstance(f, Filter) for f in filters):
                    raise ValueError('filters must be a list of Filter!')

                serialized = defaultdict(dict)

                for filter_ in filters:
                    filter_obj = serialized['pluginFilters'] if filter_.plugin_filter else serialized
                    filter_obj.update(filter_.serialize())

                json['filters'] = serialized
            else:
                json['filters'] = {}

        if voice_state is not MISSING:
            if not isinstance(voice_state, dict):
                raise ValueError('voice_state must be a dict!')

            json['voice'] = voice_state

        if not json:
            return None

        return await self.request('PATCH', f'sessions/{session_id}/players/{guild_id}',
                                  params=params, json=json)  # type: ignore

    async def destroy_player(self, guild_id: Union[str, int]) -> bool:
        """|coro|

        Destroys a player on the node.
        It's recommended that you use :func:`PlayerManager.destroy` to destroy a player.

        Returns
        -------
        bool
            Whether the player was destroyed.
        """
        session_id = self.session_id

        if not session_id:
            raise ClientError('Cannot destroy a player without a valid session ID!')

        return await self.request('DELETE', f'sessions/{session_id}/players/{guild_id}')  # type: ignore

    async def update_session(self, resuming: bool = MISSING, timeout: int = MISSING) -> Optional[RawSession]:
        """|coro|

        Update the session for this node.

        Parameters
        ----------
        resuming: bool
            Whether to enable resuming for this session.
        timeout: int
            How long the node will wait for the session to resume before destroying it, in seconds.

        Returns
        -------
        Optional[:class:`RawSession`]
            A raw response from the node containing the current session configuration, or ``None``
            if a request wasn't made due to an empty payload.
        """
        session_id = self._transport.session_id

        if not session_id:
            raise ClientError('Cannot update a session without a valid session ID!')

        json = {}

        if resuming is not MISSING:
            if not isinstance(resuming, bool):
                raise ValueError('resuming must be a bool!')

            json['resuming'] = resuming

        if timeout is not MISSING:
            if not isinstance(timeout, int) or 0 >= timeout:
                raise ValueError('timeout must be an int greater than 0!')

            json['timeout'] = timeout

        if not json:
            return None

        return await self.request('PATCH', f'sessions/{session_id}', json=json)  # type: ignore

    @overload
    async def request(self, method: str, path: str, *, to: Type[str], trace: bool = ..., versioned: bool = ..., **kwargs) -> str:
        ...

    @overload
    async def request(self, method: str, path: str, *, to: Type[T], trace: bool = ..., versioned: bool = ..., **kwargs) -> T:
        ...

    @overload
    async def request(self, method: str, path: str, *, trace: bool = ..., versioned: bool = ...,  # type: ignore
                      **kwargs) -> Union[Dict[Any, Any], List[Any], bool]:
        ...

    async def request(self,  # type: ignore
                      method: str,
                      path: str,
                      *,
                      to: Optional[Union[Type[T], str]] = None,
                      trace: bool = False,
                      versioned: bool = True,
                      **kwargs) -> Union[T, str, bool, Dict[Any, Any], List[Any]]:
        """|coro|

        .. _HTTP method: https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods

        Makes a HTTP request to this node. Useful for implementing functionality offered by plugins on the server.

        Parameters
        ----------
        method: :class:`str`
            The `HTTP method`_ for this request.
        path: :class:`str`
            The path for this request. E.g. ``sessions/{session_id}/players/{guild_id}``.
        to: Optional[Type[T]]
            The class to deserialize the response into.

            Warning
            -------
            The provided class MUST implement a classmethod called ``from_dict`` that accepts a dict or list object!

            Example:

                .. code:: python

                    @classmethod
                    def from_dict(cls, res: Union[Dict[Any, Any], List[Any]]):
                        return cls(res)
        trace: :class:`bool`
            Whether to enable trace logging for this request. This will return a more detailed error if the request fails,
            but could bloat log files and reduce performance if left enabled.
        versioned: :class:`bool`
            Whether this request should target a versioned route. For the majority of requests, this should be set to ``True``.
            This will prepend the route with the latest API version this client supports, e.g. ``v4/``.
        **kwargs: Any
            Any additional arguments that should be passed to ``aiohttp``. This could be parameters like ``json``, ``params`` etc.

        Raises
        ------
        :class:`AuthenticationError`
            If the provided authorization was invalid.
        :class:`RequestError`
            If the request was unsuccessful.
        :class:`asyncio.TimeoutError`
            If the request times out.
        :class:`aiohttp.ClientError`
            If the remote server disconnects, or a connection fails to establish etc.
        :class:`ClientError`
            A catch-all for anything not covered by the above.

        Returns
        -------
        Union[T, str, bool, Dict[Any, Any], List[Any]]
            - ``T`` or ``str`` if the ``to`` parameter was specified and either value provided.
            - The raw JSON response (``Dict[Any, Any]`` or ``List[Any]``) if ``to`` was not provided.
            - A bool, if the returned status code was ``204``. A value of ``True`` should typically mean the request was successful.
        """
        return await self._transport._request(method, path, to, trace, versioned, **kwargs)

    def __repr__(self):
        return f'<Node name={self.name} region={self.region}>'



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from .server import EndReason, Severity

if TYPE_CHECKING:
    from .abc import BasePlayer, DeferredAudioTrack
    from .node import Node
    from .server import AudioTrack


class Event:
    """ The base for all Lavalink events. """


class TrackStartEvent(Event):
    """
    This event is emitted when a track begins playing (e.g. via player.play())

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player that started to play a track.
    track: :class:`AudioTrack`
        The track that started playing.
    """
    __slots__ = ('player', 'track')

    def __init__(self, player: 'BasePlayer', track: 'AudioTrack'):
        self.player: 'BasePlayer' = player
        self.track: 'AudioTrack' = track


class TrackStuckEvent(Event):
    """
    This event is emitted when the currently playing track is stuck (i.e. has not provided any audio).
    This is typically a fault of the track's underlying audio stream, and not Lavalink itself.

    Note
    ----
    You do not need to manually trigger the start of the next track in the queue within
    this event when using the :class:`DefaultPlayer`. This is handled for you.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player associated with the stuck track.
    track: :class:`AudioTrack`
        The stuck track.
    threshold: :class:`int`
        The configured threshold, in milliseconds, after which a track is considered to be stuck
        when no audio frames were provided.
    """
    __slots__ = ('player', 'track', 'threshold')

    def __init__(self, player: 'BasePlayer', track: 'AudioTrack', threshold: int):
        self.player: 'BasePlayer' = player
        self.track: 'AudioTrack' = track
        self.threshold: int = threshold


class TrackExceptionEvent(Event):
    """
    This event is emitted when a track encounters an exception during playback.

    Note
    ----
    You do not need to manually trigger the start of the next track in the queue within
    this event when using the :class:`DefaultPlayer`. This is handled for you.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player that had the exception occur while playing a track.
    track: :class:`AudioTrack`
        The track that had the exception while playing.
    message: Optional[:class:`str`]
        The exception message.
    severity: :enum:`Severity`
        The severity of the exception.
    cause: :class:`str`
        The cause of the exception.
    """
    __slots__ = ('player', 'track', 'message', 'severity', 'cause')

    def __init__(self, player: 'BasePlayer', track: 'AudioTrack', message: Optional[str], severity: Severity,
                 cause: str):
        self.player: 'BasePlayer' = player
        self.track: 'AudioTrack' = track
        self.message: Optional[str] = message
        self.severity: Severity = severity
        self.cause: str = cause


class TrackEndEvent(Event):
    """
    This event is emitted when the player finished playing a track.

    Note
    ----
    You do not need to manually trigger the start of the next track in the queue within
    this event when using the :class:`DefaultPlayer`. This is handled for you.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player that finished playing a track.
    track: Optional[:class:`AudioTrack`]
        The track that finished playing.
        This could be ``None`` if Lavalink fails to encode the track.
    reason: :class:`EndReason`
        The reason why the track stopped playing.
    """
    __slots__ = ('player', 'track', 'reason')

    def __init__(self, player: 'BasePlayer', track: Optional['AudioTrack'], reason: EndReason):
        self.player: 'BasePlayer' = player
        self.track: Optional['AudioTrack'] = track
        self.reason: EndReason = reason


class TrackLoadFailedEvent(Event):
    """
    This is a custom event, emitted when a deferred audio track fails to
    produce a playable track. The player will not do anything by itself,
    so it is up to you to skip the broken track.

    Note
    ----
    This event will not automatically trigger the start of the next track in the queue,
    so you must ensure that you do this if you want the player to continue playing from the queue.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player responsible for playing the track.
    track: :class:`DeferredAudioTrack`
        The track that failed to produce a playable track.
    original: Optional[:class:`Exception`]
        The original error, emitted by the track.
        This may be ``None`` if the track did not raise an error,
        but rather returned ``None`` in place of a playable track.
    """
    __slots__ = ('player', 'track', 'original')

    def __init__(self, player: 'BasePlayer', track: 'DeferredAudioTrack', original: Optional[Exception]):
        self.player: 'BasePlayer' = player
        self.track: 'DeferredAudioTrack' = track
        self.original: Optional[Exception] = original


class QueueEndEvent(Event):
    """
    This is a custom event, emitted by the :class:`DefaultPlayer` when
    there are no more tracks in the queue.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player that has no more songs in queue.
    """
    __slots__ = ('player',)

    def __init__(self, player: 'BasePlayer'):
        self.player: 'BasePlayer' = player


class PlayerUpdateEvent(Event):
    """
    This event is emitted when a player's progress changes.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player that's progress was updated.
    position: :class:`int`
        The track's elapsed playback time in milliseconds.
    timestamp: :class:`int`
        The track's elapsed playback time as an epoch timestamp in milliseconds.
    connected: :class:`bool`
        Whether or not the player is connected to the voice gateway.
    ping: :class:`int`
        The player's voice connection ping.
        This is calculated from the delay between heartbeat & heartbeat ack.
        This will be -1 when the player doesn't have a voice connection.
    """
    __slots__ = ('player', 'position', 'timestamp', 'connected', 'ping')

    def __init__(self, player: 'BasePlayer', raw_state):
        self.player: 'BasePlayer' = player
        self.position: int = raw_state.get('position')
        self.timestamp: int = raw_state.get('time')
        self.connected: bool = raw_state.get('connected')
        self.ping: int = raw_state.get('ping', -1)


class NodeConnectedEvent(Event):
    """
    This is a custom event, emitted when a connection to a Lavalink node is
    successfully established.

    Attributes
    ----------
    node: :class:`Node`
        The node that was successfully connected to.
    """
    __slots__ = ('node',)

    def __init__(self, node: 'Node'):
        self.node: 'Node' = node


class NodeDisconnectedEvent(Event):
    """
    This is a custom event, emitted when the connection to a Lavalink node drops
    and becomes unavailable.

    Attributes
    ----------
    node: :class:`Node`
        The node that was disconnected from.
    code: Optional[:class:`int`]
        The status code of the event.
    reason: Optional[:class:`str`]
        The reason of why the node was disconnected.
    """
    __slots__ = ('node', 'code', 'reason')

    def __init__(self, node: 'Node', code: Optional[int], reason: Optional[str]):
        self.node: 'Node' = node
        self.code: Optional[int] = code
        self.reason: Optional[str] = reason


class NodeChangedEvent(Event):
    """
    This is a custom event, emitted when a player changes to another Lavalink node.
    Keep in mind this event can be emitted multiple times if a node disconnects and
    the load balancer moves players to a new node.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player whose node was changed.
    old_node: :class:`Node`
        The node the player was moved from.
    new_node: :class:`Node`
        The node the player was moved to.
    """
    __slots__ = ('player', 'old_node', 'new_node')

    def __init__(self, player: 'BasePlayer', old_node: 'Node', new_node: 'Node'):
        self.player: 'BasePlayer' = player
        self.old_node: 'Node' = old_node
        self.new_node: 'Node' = new_node


class NodeReadyEvent(Event):
    """
    This is a custom event, emitted when a node becomes ready.
    A node is considered ready once it receives the "ready" event from the Lavalink server.

    Attributes
    ----------
    node: :class:`Node`
        The node that became ready.
    session_id: :class:`str`
        The ID of the session.
    resumed: :class:`bool`
        Whether the session was resumed. This will be false if a brand new session was created.
    """
    __slots__ = ('node', 'session_id', 'resumed')

    def __init__(self, node: 'Node', session_id: str, resumed: bool):
        self.node: 'Node' = node
        self.session_id: str = session_id
        self.resumed: bool = resumed


class WebSocketClosedEvent(Event):
    """
    This event is emitted when an audio websocket to Discord is closed. This can happen
    happen for various reasons, an example being when a channel is deleted.

    Refer to the `Discord Developer docs <https://discord.com/developers/docs/topics/opcodes-and-status-codes#voice-voice-close-event-codes>`_
    for a list of close codes and what they mean. This event primarily exists for debug purposes,
    and no special handling of voice connections should take place unless it is absolutely necessary.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player whose audio websocket was closed.
    code: :class:`int`
        The websocket closure code.
    reason: :class:`str`
        The reason for the websocket closure.
    by_remote: :class:`bool`
        Whether the websocket was closed by the remote server.
    """
    __slots__ = ('player', 'code', 'reason', 'by_remote')

    def __init__(self, player: 'BasePlayer', code: int, reason: str, by_remote: bool):
        self.player: 'BasePlayer' = player
        self.code: int = code
        self.reason: str = reason
        self.by_remote: bool = by_remote


class IncomingWebSocketMessage(Event):
    """
    This event is emitted whenever the client receives a websocket message from the Lavalink server.

    You can use this to extend the functionality of the client, particularly useful when used with
    Lavalink server plugins that can add new HTTP routes or websocket messages.

    Attributes
    ----------
    data: Union[Dict[Any, Any], List[Any]]
        The received JSON-formatted data from the websocket.
    node: :class:`Node`
        The node responsible for this websocket message.
    """
    __slots__ = ('data', 'node')

    def __init__(self, data, node: 'Node'):
        self.data: Union[Dict[Any, Any], List[Any]] = data
        self.node: 'Node' = node


class PlayerErrorEvent(Event):
    """
    Raised when an error occurs within a :class:`BasePlayer`.

    Attributes
    ----------
    player: :class:`BasePlayer`
        The player in which the error occurred.
    original: :class:`Exception`
        The original error.
    """
    __slots__ = ('player', 'original')

    def __init__(self, player, original):
        self.player: 'BasePlayer' = player
        self.original: Exception = original
        # TODO: Perhaps an enum denoting which area of the player encountered an exception, e.g. ErrorType.PLAY.



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
from typing import (TYPE_CHECKING, Callable, Dict, Generic, Iterator, Optional,
                    Tuple, Type, TypeVar, Union, overload)

from .errors import ClientError, RequestError
from .node import Node
from .player import BasePlayer

if TYPE_CHECKING:
    from .client import Client

_log = logging.getLogger(__name__)

PlayerT = TypeVar('PlayerT', bound=BasePlayer)
CustomPlayerT = TypeVar('CustomPlayerT', bound=BasePlayer)


class PlayerManager(Generic[PlayerT]):
    """
    Represents the player manager that contains all the players.

    len(x):
        Returns the total number of stored players.
    iter(x):
        Returns an iterator of all the stored players.

    Attributes
    ----------
    client: :class:`Client`
        The Lavalink client.
    players: Dict[int, :class:`BasePlayer`]
        Cache of all the players that Lavalink has created.
    """
    __slots__ = ('client', '_player_cls', 'players')

    def __init__(self, client, player: Type[PlayerT]):
        if not issubclass(player, BasePlayer):
            raise ValueError('Player must implement BasePlayer.')

        self.client: 'Client' = client
        self._player_cls: Type[PlayerT] = player
        self.players: Dict[int, PlayerT] = {}

    def __len__(self) -> int:
        return len(self.players)

    def __iter__(self) -> Iterator[Tuple[int, PlayerT]]:
        """ Returns an iterator that yields a tuple of (guild_id, player). """
        yield from self.players.items()

    def values(self) -> Iterator[PlayerT]:
        """ Returns an iterator that yields only values. """
        yield from self.players.values()

    def find_all(self, predicate: Optional[Callable[[PlayerT], bool]] = None):
        """
        Returns a list of players that match the given predicate.

        Parameters
        ----------
        predicate: Optional[Callable[[:class:BasePlayer], bool]]
            A predicate to return specific players. Defaults to ``None``.

        Returns
        -------
        List[:class:`BasePlayer`]
            This could be a :class:`DefaultPlayer` if no custom player implementation
            was provided.
        """
        if not predicate:
            return list(self.players.values())

        return [p for p in self.players.values() if bool(predicate(p))]

    def get(self, guild_id: int) -> Optional[PlayerT]:
        """
        Gets a player from cache.

        Parameters
        ----------
        guild_id: :class:`int`
            The guild_id associated with the player to get.

        Returns
        -------
        Optional[:class:`BasePlayer`]
            This could be a :class:`DefaultPlayer` if no custom player implementation
            was provided.
        """
        return self.players.get(guild_id)

    def remove(self, guild_id: int):
        """
        Removes a player from the internal cache.

        Parameters
        ----------
        guild_id: :class:`int`
            The player to remove from cache.
        """
        if guild_id in self.players:
            player = self.players.pop(guild_id)
            player.cleanup()

    @overload
    def create(self,
               guild_id: int,
               *,
               region: Optional[str] = ...,
               endpoint: Optional[str] = ...,
               node: Optional[Node] = ...) -> PlayerT:
        ...

    @overload
    def create(self,
               guild_id: int,
               *,
               region: Optional[str] = ...,
               endpoint: Optional[str] = ...,
               node_filter: Optional[Callable[[Node], bool]] = ...) -> PlayerT:
        ...

    @overload
    def create(self,
               guild_id: int,
               *,
               region: Optional[str] = ...,
               endpoint: Optional[str] = ...,
               node: Optional[Node] = ...,
               cls: Type[CustomPlayerT]) -> CustomPlayerT:
        ...

    @overload
    def create(self,
               guild_id: int,
               *,
               region: Optional[str] = ...,
               endpoint: Optional[str] = ...,
               node_filter: Optional[Callable[[Node], bool]] = ...,
               cls: Type[CustomPlayerT]) -> CustomPlayerT:
        ...

    def create(self,
               guild_id: int,
               *,
               region: Optional[str] = None,
               endpoint: Optional[str] = None,
               node: Optional[Node] = None,
               node_filter: Optional[Callable[[Node], bool]] = None,
               cls: Optional[Type[CustomPlayerT]] = None) -> Union[CustomPlayerT, PlayerT]:
        """
        Creates a player if one doesn't exist with the given information.

        If node is provided, a player will be created on that node.
        If region is provided, a player will be created on a node in the given region.
        If endpoint is provided, Lavalink.py will attempt to parse the region from the endpoint
        and return a node in the parsed region.

        If node, region and endpoint are left unspecified, or region/endpoint selection fails,
        Lavalink.py will fall back to the node with the lowest penalty.

        Region can be omitted if node is specified and vice-versa.

        Parameters
        ----------
        guild_id: :class:`int`
            The guild_id to associate with the player.
        region: Optional[:class:`str`]
            The region to use when selecting a Lavalink node.
            Defaults to ``None``.
        endpoint: Optional[:class:`str`]
            The address of the Discord voice server.
            Defaults to ``None``.
        node: Optional[:class:`Node`]
            The node to put the player on.
            Defaults to ``None``, which selects the node with the lowest penalty.
        node_filter: Optional[Callable[[:class:`Node`], :class:`bool`]]
            A filter to use when selecting nodes this player can be assigned to.
            This cannot be used with the ``node`` parameter.
            Nodes are filtered based on the given predicate, and then again based on their penalty score.
            If no nodes are found after filtering, all available nodes will be considered without filtering.
        cls: Optional[Type[:class:`BasePlayer`]]
            The player class to use when instantiating a new player.
            Defaults to ``None`` which uses the player class provided to :class:`Client`.
            If no class was provided, this will typically be :class:`DefaultPlayer`.

            Warning
            -------
            This function could return a player of a different type to that specified in ``cls``,
            if a player was created before with a different class type.

        Raises
        ------
        :class:`ValueError`
            If the provided ``cls`` is not a valid subclass of :class:`BasePlayer`.

        Returns
        -------
        :class:`BasePlayer`
            A class that inherits ``BasePlayer``. By default, the actual class returned will
            be :class:`DefaultPlayer`, however if you have specified a custom player implementation,
            then this will be different.
        """
        if guild_id in self.players:
            return self.players[guild_id]

        cls = cls or self._player_cls  # type: ignore

        if not issubclass(cls, BasePlayer):  # type: ignore
            raise ValueError('Player must implement BasePlayer.')

        if node is not None and node_filter is not None:
            raise ValueError('node and node_filter may not be specified together')

        if node_filter is not None:
            user_filtered = [n for n in self.client.node_manager.available_nodes if node_filter(n)]

            if user_filtered:
                node = min(user_filtered, key=lambda node: node.penalty)

        if node is None and endpoint is not None:  # Prioritise endpoint over region parameter
            region = self.client.node_manager.get_region(endpoint)

        best_node = node or self.client.node_manager.find_ideal_node(region)

        if not best_node:
            raise ClientError('No available nodes!')

        id_int = int(guild_id)
        player = cls(id_int, best_node)
        self.players[id_int] = player  # type: ignore
        _log.debug('Created player with GuildId %d on node \'%s\'', id_int, best_node.name)
        return player

    async def destroy(self, guild_id: int):
        """|coro|

        Removes a player from cache, and also Lavalink if applicable.
        Ensure you have disconnected the given guild_id from the voicechannel
        first, if connected.

        Warning
        -------
        This should only be used if you know what you're doing. Players should never be
        destroyed unless they have been moved to another :class:`Node`.

        Parameters
        ----------
        guild_id: int
            The guild_id associated with the player to remove.
        """
        if guild_id in self.players:
            player = self.players.pop(guild_id)
            player.cleanup()

            if player.node:
                await player.node.destroy_player(player._internal_id)

            _log.debug('Destroyed player with GuildId %d on node \'%s\'', guild_id, player.node.name if player.node else 'UNASSIGNED')
        else:
            for node in self.client.node_manager:
                try:
                    await node.destroy_player(guild_id)
                except RequestError:  # Should never happen anyway
                    pass
